/*
 * Copyright (c) 1995, Index Data.
 * See the file LICENSE for details.
 * Sebastian Hammer, Adam Dickmeiss
 *
 * $Log: d1_grs_key.c,v $
 * Revision 1.1  1995/11/01 11:56:08  quinn
 * Added Retrieval (data management) functions en masse.
 *
 *
 */

#include <stdio.h>
#include <assert.h>
#include <ctype.h>
#include <sys/types.h>
#include <unistd.h>

#include <log.h>
#include <oid.h>

#include <recctrl.h>

#include "data1.h"

#define GRS_MAX_WORD 512

static int seqno = 0;

static void grs_init(void)
{
}

static int dumpkeys(data1_node *n, struct recExtractCtrl *p)
{
    for (; n; n = n->next)
    {
	if (n->child)
	    if (dumpkeys(n->child, p) < 0)
		return -1;
	if (n->which == DATA1N_data)
	{
	    data1_node *par = get_parent_tag(n);
	    data1_att *att = 0;
	    char *b = n->u.data.data;

	    assert(par);

	    /*
	     * cycle up towards the root until we find a tag with an att..
	     * this has the effect of indexing locally defined tags with
	     * the attribute of their ancestor in the record.
	     */
	    while (!par->u.tag.element || !(att = par->u.tag.element->att))
		if (!par->parent || !(par = get_parent_tag(par->parent)))
		    break;
	    if (!att)
		continue;

	    while (*b) /* dump the words */
	    {
		RecWord wrd;
		char buf[GRS_MAX_WORD+1];
		int i;

		(*p->init)(&wrd);      /* set defaults */
		while (*b && !isalnum(*b))
		    b++;
		if (!*b)
		    break;
		i = 0;
		while (*b && isalnum(*b))
		{
		    if (i < GRS_MAX_WORD)
			buf[i++] = *b;
		    b++;
		}
		buf[i] = '\0';
		wrd.attrSet = att->parent->ordinal;
		wrd.attrUse = att->local;
		wrd.which = Word_String;
		wrd.u.string = buf;
		wrd.seqno = seqno++;
		(*p->add)(&wrd);
	    }
	}
    }
    return 0;
}

static int grs_extract(struct recExtractCtrl *p)
{
    data1_node *n;

    seqno = 0;
    if (!(n = data1_read_record((int (*)()) p->readf, p->fd)))
	return -1;
    if (dumpkeys(n, p) < 0)
    {
	data1_free_tree(n);
	return -1;
    }
    data1_free_tree(n);
    return 0;
}

/*
 * Return: -1: Nothing done. 0: Ok. >0: Bib-1 diagnostic.
 */
static int process_comp(data1_node *n, Z_RecordComposition *c)
{
    data1_esetname *eset;
    Z_Espec1 *espec = 0;
    Z_ElementSpec *p;

    switch (c->which)
    {
	case Z_RecordComp_simple:
	    if (c->u.simple->which != Z_ElementSetNames_generic)
		return 26; /* only generic form supported. Fix this later */
	    if (!(eset = data1_getesetbyname(n->u.root.absyn,
		c->u.simple->u.generic)))
	    {
		logf(LOG_LOG, "Unknown esetname '%s'", c->u.simple->u.generic);
		return 25; /* invalid esetname */
	    }
	    espec = eset->spec;
	    break;
	case Z_RecordComp_complex:
	    if (c->u.complex->generic)
	    {
		/* insert check for schema */
		if ((p = c->u.complex->generic->elementSpec))
		    switch (p->which)
		    {
			case Z_ElementSpec_elementSetName:
			    if (!(eset = data1_getesetbyname(n->u.root.absyn,
				p->u.elementSetName)))
			    {
				logf(LOG_LOG, "Unknown esetname '%s'",
				    p->u.elementSetName);
				return 25; /* invalid esetname */
			    }
			    espec = eset->spec;
			    break;
			case Z_ElementSpec_externalSpec:
			    if (p->u.externalSpec->which == Z_External_espec1)
				espec = p->u.externalSpec->
				    u.espec1;
			    else
			    {
				logf(LOG_LOG, "Unknown external espec.");
				return 25; /* bad. what is proper diagnostic? */
			    }
			    break;
		    }
	    }
	    else
		return 26; /* fix */
    }
    if (espec)
	data1_doespec1(n, espec);
    else
	return -1;
    return 0;
}

static int grs_retrieve(struct recRetrieveCtrl *p)
{
    data1_node *node = 0, *onode = 0;
    data1_node *new;
    data1_maptab *map;
    int res, selected = 0;

    if (!(node = data1_read_record(p->readf, p->fd)))
    {
	p->diagnostic = 2;
	return 0;
    }
    if (p->score >= 0 && (new = data1_insert_taggeddata(node, node, "rank")))
    {
	new->u.data.what = DATA1I_num;
	new->u.data.data = new->u.data.lbuf;
	sprintf(new->u.data.data, "%d", p->score);
	new->u.data.len = strlen(new->u.data.data);
    }
    if ((new = data1_insert_taggeddata(node, node, "localControlNumber")))
    {
	new->u.data.what = DATA1I_text;
	new->u.data.data = new->u.data.lbuf;
	sprintf(new->u.data.data, "%d", p->localno);
	new->u.data.len = strlen(new->u.data.data);
    }

    /*
     * Does the requested format match a known schema-mapping? (this reflects
     * the overlap of schema and formatting which is inherent in the MARC
     * family)
     * NOTE: This should look at the schema-specification in the compspec
     * as well.
     */
    for (map = node->u.root.absyn->maptabs; map; map = map->next)
	if (map->target_absyn_ref == p->input_format)
	{
	    onode = node;
	    if (!(node = data1_map_record(onode, map)))
	    {
		p->diagnostic = 14;
		return 0;
	    }

	    break;
	}

    if (p->comp && (res = process_comp(node, p->comp)) > 0)
    {
	p->diagnostic = res;
	if (onode)
	    data1_free_tree(onode);
	data1_free_tree(node);
	return 0;
    }
    else if (p->comp && !res)
	selected = 1;

    switch (p->input_format != VAL_NONE ? p->input_format : VAL_SUTRS)
    {
	data1_marctab *marctab;

	case VAL_GRS1:
	    if (!(p->rec_buf = data1_nodetogr(node, selected, p->odr)))
	    {
		p->diagnostic = 2; /* this should be better specified */
		break;
	    }
	    p->rec_len = -1;
	    break;
	case VAL_SUTRS:
	    if (!(p->rec_buf = data1_nodetobuf(node, selected,
		(int*)&p->rec_len)))
	    {
		p->diagnostic = 2;
		break;
	    }
	    break;
	default:
	    for (marctab = node->u.root.absyn->marc; marctab;
		marctab = marctab->next)
		if (marctab->reference == p->input_format)
		    break;
	    if (!marctab)
	    {
		p->diagnostic = 227;
		break;
	    }
	    if (!(p->rec_buf = data1_nodetomarc(marctab, node, selected,
		(int*)&p->rec_len)))
	    {
		p->diagnostic = 2;
		break;
	    }
    }
    if (node)
	data1_free_tree(node);
    if (onode)
	data1_free_tree(onode);
    return 0;
}

static struct recType grs_type =
{
    "grs",
    grs_init,
    grs_extract,
    grs_retrieve
};

RecType recTypeGrs = &grs_type;
